#include <iostream>
#include <string>
#include <vector>
#include <cstring>
#include <unistd.h>

/*
    This program has the purpose of explaining an X11 modeline.
*/

using namespace std;

struct Modeline
{
    string name;
    double clock;
    struct { unsigned active, bstart, bend, length; bool negative; } horiz, vert;
    
    bool doublescan, interlace;
    
    // Default sync polarities (defined by some standard):
    //    0..399 lines: +H -V
    //  400..479 lines: -H +V
    //  480..767 lines: -H -V
    //  768..    lines: +H +V
    bool defaulthsyncpolarity() const { return horiz.active >= 400 && horiz.active < 768; }
    bool defaultvsyncpolarity() const { return vert.active < 400 || (vert.active >= 480 && vert.active < 768); }
    
    const double getvrefresh() const
    {
        // Number of times the screen is refreshed per second
        double res = clock * 1e6 / (horiz.length * vert.length);
        if(doublescan)res /= 2;
        return res;
    }
    const double gethsync() const
    {
        // Number of scanlines handled per second
        double res = getvrefresh() * vert.length;
        if(doublescan)res += res;
        return res;
    }
    
    void Explain() const
    {
        // Number of times the screen is refreshed per second:
        const double vrefresh = getvrefresh();
        // Number of scanlines handled per second:
        const double hsync = gethsync();
        
        // Length of a screen update in seconds
        const double screenlength = 1.0 / vrefresh;
        
        // Length of a scanline in seconds
        const double scanlinelength = screenlength / vert.length;
        
        // Blank sizes
        const unsigned right  = horiz.bstart - horiz.active;
        const unsigned left   = horiz.length - horiz.bend;
        const unsigned bottom = vert.bstart - vert.active;
        const unsigned top    = vert.length - vert.bend;
        
        // Length (in seconds) of hsync:
        const double hslen = scanlinelength * (horiz.bend-horiz.bstart) / horiz.length;
        // Length (in seconds) of vsync:
        const double vslen = screenlength  * (vert.bend-vert.bstart) / vert.length;
        
        cout << "Modeline \"" << name << "\"  " << clock << "  "
                     << horiz.active << ' ' << horiz.bstart << ' ' << horiz.bend << ' ' << horiz.length
             << "  " << vert.active << ' ' << vert.bstart << ' ' << vert.bend << ' ' << vert.length;
        if(doublescan)cout << " Doublescan";
        if(interlace)cout << " Interlace";

        if(horiz.negative != defaulthsyncpolarity())
            cout << ' ' << (horiz.negative ? '-' : '+') << "HSync";
        if(vert.negative != defaultvsyncpolarity())
            cout << ' ' << (vert.negative ? '-' : '+') << "VSync";
        cout << endl;

        cout << "#  Resolution: " << horiz.active << 'x' << vert.active
                                  << " @ " << vrefresh << " Hz"
                                     " - pixclock " << (1e6/clock) << " ps\n"
                "#  hsync " << (hsync/1000) << " kHz, vrefresh " << vrefresh << " Hz\n"
                "#  Blank parts: " << left << " left, "
                                   << right << " right, "
                                   << top << " top, "
                                   << bottom << " bottom\n"

                "#  Sync times: " << (1e6*hslen) << " us horiz (" << (horiz.bend-horiz.bstart) << " px), "
                                  << (1e6*vslen) << " us vert (" << (vert.bend-vert.bstart) << " px)\n";
        if(doublescan) cout << "#  Mode is doublescan mode (THUS THINGS MAY BE WRONG).\n";
        if(interlace) cout << "#  Mode is interlaced (THUS THINGS MAY BE WRONG).\n";
        cout << "#  " << (horiz.negative ? '-' : '+') << "HSync, "
                      << (vert.negative ? '-' : '+') << "VSync\n";
        
        for(unsigned a=0; a<3; ++a)
        {
            static const char *const expl[3] = {"simple", "working", "exact"};
            bool doposition = a>0;
            bool dotimings = a>1;
            
            cout << "#  modetest commandline (" << expl[a] << "): modetest";
            cout << " -x" << horiz.active << '+' << (left+right);
            if(doposition)
                cout << '/' << (left * 100 / (left+right));
            cout << " -y" << vert.active << '+' << (top+bottom);
            if(doposition)
                cout << '/' << (top * 100 / (top+bottom));
            cout << " -r" << (unsigned)(vrefresh+0.5);
            if(dotimings)
            {
                cout << " -X" << hslen << " -Y" << vslen;
            }
            cout << "\n";
        }
        
        cout << endl;
    }
};

const vector<string> paloiksi(const string &s)
{
    vector<string> result;
    unsigned b=s.size();
    bool laina=false;
    string part;
    for(unsigned a=0; a<b; ++a)
    {
        if(s[a]==' ' && !laina)
        {
            if(part.size())
            {
                result.push_back(part);
                part = "";
            }
            continue;
        }
        if(s[a]=='"') { laina=!laina; continue; }
        part += s[a];
    }
    if(part.size()) result.push_back(part);
    return result;
}

int main(void)
{
    if(isatty(0))
    {
        cerr << "This program explains modelines.\n"
                "Copyright (C) 1992,2003 Bisqwit (http://iki.fi/bisqwit/)\n"
                "Please copypaste a valid XF86 modeline here:" << endl;
    }
    while(true)
    {
        string s;
        getline(cin, s);
        if(!cin.good())break;
        
        unsigned hashpos=s.find('#');
        if(!hashpos)continue; // line is a comment
        if(hashpos != s.npos)
            s = s.substr(0, hashpos);
        
        vector<string> parts = paloiksi(s);
        
        vector<string> errors;
        
        if(parts.size() < 11)
            errors.push_back("missing parameters in modeline");
        if(parts.size() > 0
        && strcasecmp(parts[0].c_str(), "modeline"))
            errors.push_back("first word should be 'modeline'");
        
        Modeline line;
        if(parts.size() > 1)line.name = parts[1];
        if(parts.size() > 2)line.clock = strtod(parts[2].c_str(), NULL);
        if(parts.size() > 3)line.horiz.active = strtol(parts[3].c_str(), NULL, 10);
        if(parts.size() > 4)line.horiz.bstart = strtol(parts[4].c_str(), NULL, 10);
        if(parts.size() > 5)line.horiz.bend = strtol(parts[5].c_str(), NULL, 10);
        if(parts.size() > 6)line.horiz.length = strtol(parts[6].c_str(), NULL, 10);
        if(parts.size() > 7)line.vert.active = strtol(parts[7].c_str(), NULL, 10);
        if(parts.size() > 8)line.vert.bstart = strtol(parts[8].c_str(), NULL, 10);
        if(parts.size() > 9)line.vert.bend = strtol(parts[9].c_str(), NULL, 10);
        if(parts.size() >10)line.vert.length = strtol(parts[10].c_str(), NULL, 10);
        
        line.doublescan = false;
        line.interlace = false;
        line.horiz.negative = line.defaulthsyncpolarity();
        line.vert.negative  = line.defaultvsyncpolarity();
        for(unsigned a=11; a<parts.size(); ++a)
        {
            if(!strcasecmp(parts[a].c_str(), "doublescan")) line.doublescan=true;
            else if(!strcasecmp(parts[a].c_str(), "interlace")) line.interlace=true;
            else if(!strcasecmp(parts[a].c_str(), "-hsync")) line.horiz.negative = true;
            else if(!strcasecmp(parts[a].c_str(), "+hsync")) line.horiz.negative = false;
            else if(!strcasecmp(parts[a].c_str(), "-vsync")) line.vert.negative = true;
            else if(!strcasecmp(parts[a].c_str(), "+vsync")) line.vert.negative = false;
            else errors.push_back("whatever does '"+parts[a]+"' mean");
        }
        
        bool checkonly = true;
        
        if(errors.size())
        {
            cerr << "Errors detected in '" << s << "':\n";
            for(unsigned a=0; a<errors.size(); ++a)
                cerr << " - " << errors[a] << endl;
        }
        else
        {
            if(checkonly)
            {
                const double v = line.getvrefresh();
                const double h = line.gethsync();
                if(v >= 43 && v <= 90
                && h >= 22e3 && h <= 49e3)
                {
                    cout << s << endl;
                }
                else
                {
                    cout << "# HS " << (h/1000) << " kHz, VR " << v << "\n"
                            "#" << s << endl;
                }
            }
            line.Explain();
        }
    }
    return 0;
}
